# This file was auto-generated by Fern from our API Definition.

import typing

FolderEntitiesListRequestEntityStatus = typing.Union[
    typing.Literal["ACTIVE", "ARCHIVED", "PENDING_DELETION"], typing.Any
]
